program test

#include "petsc/finclude/petscvec.h"
#include "petsc/finclude/petscdmplex.h"
#include "petsc/finclude/petscdmlabel.h"
 use petscvec
 use petscdmplex
 use petscsys

 implicit none

 DM                              :: dm_dumm, dm_geom
 Vec                             :: vglo, vloc
 PetscMPIInt                     :: myid, nproc
 PetscErrorCode                  :: ierr
 PetscSection                    :: section
 PetscInt                        :: dim
 PetscSF                         :: sf
 PetscReal, target, dimension(1) :: sol_tmp
 PetscReal, pointer              :: pSol(:) !=> sol_tmp
 PetscViewer                     :: viewer
 character(100)                  :: fname
 integer                         :: i
 integer                         :: vst, vend
 integer                         :: nnode


 call PetscInitialize(PETSC_NULL_CHARACTER, ierr)
 if(ierr.ne.0) then
   write(*,*) 'Fail to initialize PETSc ...'
   stop
 endif
 call MPI_COMM_RANK(MPI_COMM_WORLD, myid, ierr)
 call MPI_COMM_SIZE(MPI_COMM_WORLD, nproc, ierr)
 write(*,*) 'MPI information ... ', myid, nproc

 fname = './square.msh'
 call DMPlexCreateGmshFromFile(PETSC_COMM_WORLD, fname, PETSC_TRUE, dm_dumm, ierr);CHKERRA(ierr)
 call DMPlexDistributeSetDefault(dm_dumm, PETSC_FALSE, ierr);CHKERRA(ierr)
 if(nproc.gt.1) then
   call DMPlexDistribute(dm_dumm, 0, PETSC_NULL_SF, dm_geom, ierr);CHKERRA(ierr)
 else
   call DMClone(dm_dumm, dm_geom, ierr);CHKERRA(ierr)
 endif

 call DMGetDimension(dm_geom, dim, ierr);CHKERRA(ierr)

 ! define section with 1-dof on vertex for "vertex-dm"
 call PetscSectionCreate(PETSC_COMM_WORLD, section, ierr);CHKERRA(ierr)
 call PetscSectionSetNumFields(section, 1, ierr);CHKERRA(ierr)
 call PetscSectionSetFieldComponents(section, 0, 1, ierr);CHKERRA(ierr)
 call DMPlexGetDepthStratum(dm_geom, 0, vst, vend, ierr);CHKERRA(ierr) ! vertex
 call PetscSectionSetChart(section, vst, vend, ierr);CHKERRA(ierr)
 do i=vst,vend-1
   call PetscSectionSetDof(section, i, 1, ierr);CHKERRA(ierr)
   call PetscSectionSetFieldDof(section, i, 0, 1, ierr);CHKERRA(ierr)
 enddo
 call PetscSectionSetUp(section, ierr);CHKERRA(ierr)
 call DMSetLocalSection(dm_geom, section, ierr);CHKERRA(ierr)
 call PetscSectionDestroy(section, ierr);CHKERRA(ierr)
 ! setup star forest using that section
 call DMGetSectionSF(dm_geom, sf, ierr);CHKERRA(ierr)
 call PetscSFSetUp(sf, ierr);CHKERRA(ierr)

 ! get coordinates over vertex 
 call DMPlexGetDepthStratum(dm_geom, 0, vst, vend, ierr);CHKERRA(ierr)
 nnode = vend - vst ! number of vertex each proc has

 ! connect pointer and target
 pSol => sol_tmp

 ! borrow local and global vectors
 call DMGetLocalVector(dm_geom, vloc, ierr);CHKERRA(ierr)
 call DMGetGlobalVector(dm_geom, vglo, ierr);CHKERRA(ierr)


 ! Perform halo exchange

 !=====================================================================!
! ! Version - 1: "Local and Global"; No problem
! do i=1,nnode
!   pSol = myid + 1
!   call VecSetValues(vloc, 1, i-1, pSol, INSERT_VALUES, ierr);CHKERRA(ierr)
! enddo
!
! ! halo exchange
! call VecZeroEntries(vglo, ierr);CHKERRA(ierr)
! call DMLocalToGlobalBegin(dm_geom, vloc, ADD_VALUES, vglo, ierr);CHKERRA(ierr)
! call DMLocalToGlobalEnd(dm_geom, vloc,   ADD_VALUES, vglo, ierr);CHKERRA(ierr)
! call DMGlobalToLocalBegin(dm_geom, vglo, INSERT_VALUES, vloc, ierr);CHKERRA(ierr)
! call DMGlobalToLocalEnd(dm_geom, vglo, INSERT_VALUES, vloc, ierr);CHKERRA(ierr)
 !=====================================================================!

 !=====================================================================!
 ! Version - 2: "Local and Local"; Not working
 do i=1,nnode
   pSol = myid + 1
   call VecSetValues(vloc, 1, i-1, pSol, INSERT_VALUES, ierr);CHKERRA(ierr)
 enddo

 ! halo exchange
 call DMLocalToLocalBegin(dm_geom, vloc, ADD_VALUES, vloc, ierr);CHKERRA(ierr)
 call DMLocalToLocalEnd(dm_geom, vloc,   ADD_VALUES, vloc, ierr);CHKERRA(ierr)
 !=====================================================================!


 ! create viewer
 call PetscViewerCreate(PETSC_COMM_WORLD, viewer, ierr);CHKERRA(ierr)
 
 ! set viewer type
 call PetscViewerSetType(viewer, PETSCVIEWERVTK, ierr);CHKERRA(ierr)
 
 ! set viewer
 fname = './output.vtu'
 call PetscViewerFileSetName(viewer, fname, ierr);CHKERRA(ierr) 
 
 ! print dm object to the file
 call VecView(vloc, viewer, ierr);CHKERRA(ierr)
 
 ! destroy
 call PetscViewerDestroy(viewer, ierr);CHKERRA(ierr)

 ! return borrowed vectors
 call DMRestoreLocalVector(dm_geom, vloc, ierr);CHKERRA(ierr)
 call DMRestoreGlobalVector(dm_geom, vglo, ierr);CHKERRA(ierr)

 ! destroy and free memory
 call DMDestroy(dm_dumm, ierr);CHKERRA(ierr)
 call DMDestroy(dm_geom, ierr);CHKERRA(ierr)

 write(*,*) 'done ...', myid
 call PetscFinalize(ierr)

 stop
end program test
